{% extends "wbcore/email_base_template.html" %}

{% block body %}
<p>
    <strong>A Customer Trade was saved that might be related to your managed sub accounts:</strong><br />
</p>
<p style="text-align: left; text-decoration: underline;">Claim info:</p>
<p style="text-align: left;">
    <ul style="text-align: left;">
        <li><b>Bank:</b> {{model.bank}}</li>
        <li><b>Product:</b> {{model.product.title}}</li>
        <li><b>Trade Date:</b> {{model.transaction_date}}</li>
        <li><b>Shares:</b> {{model.shares}}</li>
        <li><b>Price:</b> {{model.price}}</li>
        {% if notification.external_identifier %}
        <li><b>External Identifier:</b> {{model.external_id}}</li>
        {% endif %}
        {% if notification.message %}
        <li><b>Related SubAccounts:</b> {{notification.message}}</li>
        {% endif %}
        {% if model.comment%}
        <li><b>Comment:</b> <i>{{model.comment|safe}}</i></li>
        {% endif %}
      </ul>
</p>
{% endblock %}
