# Generated by Django 5.0.6 on 2024-08-15 13:59

from decimal import Decimal

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("wbportfolio", "0064_alter_portfolio_managers_portfolio_is_tracked_and_more"),
    ]

    operations = [
        migrations.AlterModelManagers(
            name="portfolio",
            managers=[],
        ),
        migrations.AddField(
            model_name="claim",
            name="as_shares",
            field=models.BooleanField(default=True),
        ),
        migrations.AlterField(
            model_name="accountreconciliationline",
            name="nominal_value",
            field=models.DecimalField(
                decimal_places=4,
                default=Decimal("0"),
                help_text="The nominal value computed through the Workbench",
                max_digits=18,
            ),
        ),
        migrations.AlterField(
            model_name="accountreconciliationline",
            name="nominal_value_external",
            field=models.DecimalField(
                decimal_places=4,
                default=Decimal("0"),
                help_text="The nominal value externally provided through the Account holder. Initially set to the number of shares computed through the Workbench",
                max_digits=18,
            ),
        ),
        migrations.AlterField(
            model_name="accountreconciliationline",
            name="shares",
            field=models.DecimalField(
                decimal_places=4,
                default=Decimal("0"),
                help_text="The number of shares computed through the Workbench",
                max_digits=18,
            ),
        ),
        migrations.AlterField(
            model_name="accountreconciliationline",
            name="shares_external",
            field=models.DecimalField(
                decimal_places=4,
                default=Decimal("0"),
                help_text="The number of shares externally provided through the Account holder. Initially set to the number of shares computed through the Workbench",
                max_digits=18,
            ),
        ),
        migrations.AlterField(
            model_name="claim",
            name="date",
            field=models.DateField(verbose_name="Trade Date"),
        ),
        migrations.AlterField(
            model_name="claim",
            name="reference",
            field=models.CharField(blank=True, max_length=255, null=True, verbose_name="Additional Reference"),
        ),
    ]
