# Generated by Django 4.1.11 on 2023-10-04 06:48

import django_fsm
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("wbportfolio", "0047_remove_product_default_sub_account"),
    ]

    operations = [
        migrations.AlterField(
            model_name="trade",
            name="status",
            field=django_fsm.FSMField(
                choices=[
                    ("DRAFT", "Draft"),
                    ("SUBMIT", "Submit"),
                    ("EXECUTED", "Executed"),
                    ("CONFIRMED", "Confirmed"),
                    ("FAILED", "Failed"),
                ],
                default="CONFIRMED",
                max_length=50,
                verbose_name="Status",
            ),
        ),
    ]
